///////////////////////////////////////////////////////////////////////////////
//
//	rio.cpp
//
///////////////////////////////////////////////////////////////////////////////
#include	<stdio.h>
#include	<string.h>
#include	<time.h>
#include	<sys/stat.h>
#include	"std.h"
#include	"binary.h"
#include	"rio.h"

// platform dependencies
#if defined(_WIN32)
	// MS VC++ v5.0
	#include	<conio.h>
	#define		OUTPORT( p, v )			_outp( p, v )
	#define		INPORT( p )				_inp( p )
	#define		CLOCK_SECOND			1000
#elif defined(__linux__)
	// linux g++
	#include	<unistd.h>
//	#include	<sys/perm.h>
	#include	<asm/io.h>
	#define		OUTPORT(p,v)			outb( v, p )
	#define		INPORT(p)				inb( p )
	#define		CLOCK_SECOND			1000000
#else
	// not supported
	#error ! ! compiler/platform not supported ! !
#endif

// port offset constants
#define		OFFSET_PORT_DATA		0
#define		OFFSET_PORT_STATUS		1
#define		OFFSET_PORT_CONTROL		2

// max tx/rx block retry
#define		MAX_RETRY				3

// delay's
#define		IODELAY(c)				{ for( int _iA=0; _iA<c; ++_iA ) INPORT(m_iPortStatus); }
#define		DELAY(t)				{ long _lTime=clock()+t; while( _lTime > clock() ) ; }

// new, delete
#define		NEWBLOCK( p )			{ p = new UCHAR[ CRIO_SIZE_32KBLOCK ]; if ( !p ) { ERRORSTR( "new failed\n" ); return FALSE; } }
#define		ZERONEWBLOCK( p )		{ NEWBLOCK(p); memset(p, 0, CRIO_SIZE_32KBLOCK); }
#define		DELETEBLOCK( p )		{ if ( p ) { delete[] p; p = NULL; } }

// output
#define		INFOSTR					printf
#define		ERRORSTR				INFOSTR

// 32Kblock used flag
#define		ID_32KBLOCK_USED		0x00
#define		ID_32KBLOCK_FREE		0xff

// command out
#define		COMMANDOUT(v1, v2, v3)	{ OUTPORT(m_iPortData, v1); OUTPORT(m_iPortControl, v2); OUTPORT(m_iPortControl, v3); }
// wait for reply
#define		WAITNIBBLE( v1 )		{ if (!WaitInput(v1)) return FALSE; }
#define		WAITACK()				{ if (!WaitAck()) return FALSE; }

// io delay constants
#define		IODELAY_1				20000
#define		IODELAY_2				20

///////////////////////////////////////////////////////////////////////////////
// return file only
static char* GetFile( char* pszPathFile )
{
	int iLength = strlen( pszPathFile );
	if ( !iLength )
		return pszPathFile;

	char* pc = pszPathFile + iLength - 1;
	while( *pc != '\\' && *pc != '/' && *pc != ':' )
	{
		if ( pc == pszPathFile )
			return pc;
		--pc;
	}
	++pc;

	return pc;
}

///////////////////////////////////////////////////////////////////////////////
// set/unset, constructors and destructors
void CRio::Unset( void )
{
}

BOOL CRio::Set( int iPortBase )
{
	// cleanup any previous Set()
	Unset();

	// determine ports
	m_iPortBase = iPortBase;
	m_iPortData = m_iPortBase + OFFSET_PORT_DATA;
	m_iPortStatus = m_iPortBase + OFFSET_PORT_STATUS;
	m_iPortControl = m_iPortBase + OFFSET_PORT_CONTROL;

	// if linux, request access to required ports
	#if defined(__linux__)
		if ( ioperm(m_iPortBase, 3, 1) )
		{
			ERRORSTR( "access to IO ports denied\n" );
			return FALSE;
		}
	#endif

	// flag as set ok
	m_bSet = TRUE;

	return TRUE;
}

CRio::CRio()
{
	m_bSet = FALSE;
}

CRio::~CRio()
{
	Unset();
}

///////////////////////////////////////////////////////////////////////////////
// find first free 32K block
UINT CRio::FindFirstFree32KBlock( void )
{
	USHORT uiA;
	UCHAR* puc = m_cDirBlock.m_auc32KBlockUsed;
	for( uiA=0; uiA<CRIO_MAX_32KBLOCK; ++uiA, ++puc )
	{
		if ( *puc == ID_32KBLOCK_FREE )
			return uiA;
	}
	return 0xffff;
}

// calculate checksum1 (checksum of directory header)
UINT CRio::CalculateChecksum1( void )
{
	USHORT usChecksum = m_cDirBlock.m_cDirHeader.m_usChecksum1;
	USHORT* paus = (USHORT*)&m_cDirBlock.m_cDirHeader;
	for( int iA=0; iA<(sizeof(CDirHeader)/sizeof(short)); ++iA )
		usChecksum -= *paus++;
	return usChecksum;
}

// calculate checksum2 (checksum of directory entries, used flags and FAT)
UINT CRio::CalculateChecksum2( void )
{
	USHORT usChecksum = 0;
	USHORT* paus = (USHORT*)m_cDirBlock.m_acDirEntry;
	int iSize = ( sizeof(CDirBlock) - sizeof(CDirHeader) ) / sizeof(short);
	for( int iA=0; iA<iSize; ++iA )
		usChecksum -= *paus++;
	return usChecksum;
}

// wait for	requested input from status port
BOOL CRio::WaitInput( int iValue )
{
	long lTime = clock() + CLOCK_SECOND;
	UCHAR ucRx;
	while( lTime > clock() )
	{
		ucRx = INPORT( m_iPortStatus ) & 0xf8;

		if ( ucRx == (int)iValue )
			return TRUE;
	}

	ERRORSTR( "WaitInput() failed, 0x%02x <> 0x%02x\n", ucRx, iValue );

	return FALSE;
}

// wait for	ack
BOOL CRio::WaitAck( void )
{
	long lTime = clock() + CLOCK_SECOND;
	while( lTime > clock() )
	{
		UCHAR ucRx = INPORT( m_iPortStatus );
		if ( ucRx & 0x08 )
			return TRUE;
	}

	ERRORSTR( "WaitAck() failed\n" );

	return FALSE;
}


// io intro
BOOL CRio::IOIntro( void )
{
	OUTPORT( m_iPortControl, B_00000100 );

	COMMANDOUT( B_10101000, B_00001100, B_00000100 );

	OUTPORT( m_iPortControl, B_00000000 );
	IODELAY( IODELAY_1 );

	OUTPORT( m_iPortControl, B_00000100 );
	IODELAY( IODELAY_1 );

	COMMANDOUT( B_10101101, B_00001100, B_00000100 );
	COMMANDOUT( B_01010101, B_00000000, B_00000100 );
	COMMANDOUT( B_10101110, B_00001100, B_00000100 );
	COMMANDOUT( B_10101010, B_00000000, B_00000100 );
	COMMANDOUT( B_10101000, B_00001100, B_00000100 );

	OUTPORT( m_iPortControl, B_00000000 );
	IODELAY( IODELAY_1 );

	OUTPORT( m_iPortControl, B_00000100 );
	IODELAY( IODELAY_1 );

	return TRUE;
}

// io outro
BOOL CRio::IOOutro( void )
{
	COMMANDOUT( B_10101101, B_00001100, B_00000100 );
	COMMANDOUT( B_11111111, B_00000000, B_00000100 );
	OUTPORT( m_iPortData, B_00000000 );
	return TRUE;
}

// tx 32K block retry
BOOL CRio::Tx32KBlockRetry( void* pv, UINT uiPos32KBlock, UINT uiPos32KBlockPrev,
	UINT uiPos32KBlockNext )
{
	int iA, iB;

	// io intro
	if ( !IOIntro() )
		return FALSE;

	// prepare 32K block
	for( iA=0; iA<4; ++iA )
	{
		UINT uiPos512ByteBlock = (uiPos32KBlock * 32768 + iA * 8192) / 512;
		UINT uiPosHi = uiPos512ByteBlock / 16384;
		UINT uiPosMid = (uiPos512ByteBlock & 0x3f00) >> 8;
		UINT uiPosLo = uiPos512ByteBlock & 0xff;

		COMMANDOUT( B_10101011, B_00001100, B_00000100 );
		COMMANDOUT( uiPosHi, B_00000000, B_00000100 );
		COMMANDOUT( B_10100001, B_00001100, B_00000100 );
		COMMANDOUT( B_01100000, B_00000000, B_00000100 );
		COMMANDOUT( B_10100010, B_00001100, B_00000100 );
		COMMANDOUT( uiPosLo, B_00000000, B_00000100 );
		COMMANDOUT( uiPosMid, B_00000000, B_00000100 );
		COMMANDOUT( B_10100001, B_00001100, B_00000100 );
		COMMANDOUT( B_11010000, B_00000000, B_00000100 );

		WAITACK();

		COMMANDOUT( B_10100001, B_00001100, B_00000100 );
		COMMANDOUT( B_01110000, B_00000000, B_00000100 );
		COMMANDOUT( B_10100000, B_00001100, B_00000100 );

		OUTPORT( m_iPortControl, B_00000000 );
		IODELAY( 100 );

		OUTPORT( m_iPortControl, B_00000100 );
		IODELAY( 100 );
	}

	// send 32K in 512 byte chunks
	UCHAR* pauc = (UCHAR*)pv;
	for( iA=0; iA<(32768/512); ++iA, pauc+=512 )
	{
		UINT uiPos512ByteBlock = (uiPos32KBlock * 32768 + iA * 512) / 512;
		UINT uiPosHi = uiPos512ByteBlock / 16384;
		UINT uiPosMid = (uiPos512ByteBlock & 0x3f00) >> 8;
		UINT uiPosLo = uiPos512ByteBlock & 0xff;

		// issue upload 512 byte block command
		COMMANDOUT( B_10101011, B_00001100, B_00000100 );
		COMMANDOUT( uiPosHi, B_00000000, B_00000100 );
		COMMANDOUT( B_10100001, B_00001100, B_00000100 );
		COMMANDOUT( B_10000000, B_00000000, B_00000100 );
		COMMANDOUT( B_10100010, B_00001100, B_00000100 );
		COMMANDOUT( B_00000000, B_00000000, B_00000100 );
		COMMANDOUT( uiPosLo, B_00000000, B_00000100 );
		COMMANDOUT( uiPosMid, B_00000000, B_00000100 );

		COMMANDOUT( B_10100011, B_00001100, B_00000100 );

		// create checksum of 512 byte block
		USHORT usChecksum = 0;
		USHORT* paus = (USHORT*)pauc;
		for( iB=0; iB<(512/sizeof(short)); ++iB, ++paus )
			usChecksum -= *paus;

		// clock out data
		UCHAR* pauc2 = pauc;
		for( iB=0; iB<512; ++iB, ++pauc2 )
		{
			OUTPORT( m_iPortData, (*pauc2) );

			if ( !(iB & 1) )
				OUTPORT( m_iPortControl, B_00000000 );
			else
				OUTPORT( m_iPortControl, B_00000100 );

			IODELAY( 1 );
		}

		// prepare end of block
		CEnd512ByteBlock cEnd512ByteBlock;
		memset( &cEnd512ByteBlock, 0, sizeof(CEnd512ByteBlock) );
		cEnd512ByteBlock.m_usChecksum = usChecksum;
		if ( uiPos32KBlockNext == 0xffff )
		{
			cEnd512ByteBlock.m_ulPos512ByteBlockNextMult256 = 0xffffffff;
			cEnd512ByteBlock.m_ucPos8192KBlockNext1 = 0xff;
			cEnd512ByteBlock.m_ucPos8192KBlockNext2 = 0xff;
		}
		else
		{
			cEnd512ByteBlock.m_ulPos512ByteBlockNextMult256 = (uiPos32KBlockNext * 64) * 256;
			cEnd512ByteBlock.m_ucPos8192KBlockNext1 = uiPos32KBlockNext / 256;
			cEnd512ByteBlock.m_ucPos8192KBlockNext2 = uiPos32KBlockNext / 256;
		}

		if ( uiPos32KBlockPrev == 0xffff )
		{
			cEnd512ByteBlock.m_ulPos512ByteBlockPrevMult256 = 0xffffffff;
			cEnd512ByteBlock.m_ucPos8192KBlockPrev1 = 0xff;
			cEnd512ByteBlock.m_ucPos8192KBlockPrev2 = 0xff;
			cEnd512ByteBlock.m_usPos32KBlockPrevMult256 = 0xffff;
		}
		else
		{
			cEnd512ByteBlock.m_ulPos512ByteBlockPrevMult256 = (uiPos32KBlockPrev * 64) * 256;
			cEnd512ByteBlock.m_ucPos8192KBlockPrev1 = uiPos32KBlockPrev / 256;
			cEnd512ByteBlock.m_ucPos8192KBlockPrev2 = uiPos32KBlockPrev / 256;
			cEnd512ByteBlock.m_usPos32KBlockPrevMult256 = uiPos32KBlockPrev * 256;
		}

		// output end of block
		pauc2 = (UCHAR*)&cEnd512ByteBlock;
		for( iB=0; iB<sizeof(CEnd512ByteBlock); ++iB, ++pauc2 )
		{
			OUTPORT( m_iPortData, (*pauc2) );

			if ( !(iB & 1) )
				OUTPORT( m_iPortControl, B_00000000 );
			else
				OUTPORT( m_iPortControl, B_00000100 );

			IODELAY( 1 );
		}

		// end of tx
		COMMANDOUT( B_10100001, B_00001100, B_00000100 );
		COMMANDOUT( B_00010000, B_00000000, B_00000100 );

		WAITACK();

		COMMANDOUT( B_10100001, B_00001100, B_00000100 );
		COMMANDOUT( B_01110000, B_00000000, B_00000100 );
		COMMANDOUT( B_10100000, B_00001100, B_00000100 );

		OUTPORT( m_iPortControl, B_00000000 );
		IODELAY( 100 );

		OUTPORT( m_iPortControl, B_00000100 );
		IODELAY( 100 );
	}

	return TRUE;
}

// tx 32K block
BOOL CRio::Tx32KBlock( void* pv, UINT uiPos32KBlock, UINT uiPos32KBlockPrev,
	UINT uiPos32KBlockNext )
{
	int iRetry = 0;
	while( iRetry < MAX_RETRY )
	{
		if ( Tx32KBlockRetry(pv, uiPos32KBlock, uiPos32KBlockPrev, uiPos32KBlockNext) )
			return TRUE;

		DELAY( CLOCK_SECOND );
		++iRetry;
	}

	ERRORSTR( "too many retries for Tx32KBlock()\n" );

	return FALSE;
}

// rx 32K block retry
BOOL CRio::Rx32KBlockRetry( void* pv, UINT uiPos32KBlock )
{
	int iA, iB;

	// io intro
	if ( !IOIntro() )
		return FALSE;

	// get 32K in 512 byte chunks
	UCHAR* pauc = (UCHAR*)pv;
	for( iA=0; iA<(32768/512); ++iA, pauc+=512 )
	{
		UINT uiPos512ByteBlock = (uiPos32KBlock * 32768 + iA * 512) / 512;
		UINT uiPosHi = uiPos512ByteBlock / 16384;
		UINT uiPosMid = (uiPos512ByteBlock & 0x3f00) >> 8;
		UINT uiPosLo = uiPos512ByteBlock & 0xff;

		// issue download 512 byte block command
		COMMANDOUT( B_10101011, B_00001100, B_00000100 );
		COMMANDOUT( uiPosHi, B_00000000, B_00000100 );
		COMMANDOUT( B_10100001, B_00001100, B_00000100 );
		COMMANDOUT( B_00000000, B_00000000, B_00000100 );
		COMMANDOUT( B_10100010, B_00001100, B_00000100 );
		COMMANDOUT( B_00000000, B_00000000, B_00000100 );
		COMMANDOUT( uiPosLo, B_00000000, B_00000100 );
		COMMANDOUT( uiPosMid, B_00000000, B_00000100 );

		if ( INPORT(m_iPortStatus) & 0x08 )
		{
			ERRORSTR( "Rx32KBlockRetry() failed\n" );
			return FALSE;
		}

		WAITACK();

		COMMANDOUT( B_10100000, B_00001100, B_00000100 );

		// clock in data (store first 512 bytes, ignore rest)
		UCHAR ucIn, ucRx;
		UCHAR* pauc2 = pauc;
		for( iB=0; iB<(512+16); ++iB )
		{
			// get hi nibble
			OUTPORT( m_iPortControl, B_00000000 );
			IODELAY( IODELAY_2 );
			ucRx = INPORT( m_iPortStatus );
			ucIn = ((ucRx & 0xf0) ^ 0x80) >> 4;

			// get lo nibble and combine with previous nibble to make byte
			OUTPORT( m_iPortControl, B_00000100 );
			IODELAY( IODELAY_2 );
			ucRx = INPORT( m_iPortStatus );
			ucIn |= (ucRx & 0xf0) ^ 0x80;

			// if first 512 bytes
			if ( iB < 512 )
			{
				// reverse bits in byte
				UCHAR ucReversed = 0;
				for( int iC=0; iC<8; ++iC )
				{
					ucReversed <<= 1;
					ucReversed |= (ucIn & 1);
					ucIn >>= 1;
				}

				// store reversed byte
				*pauc2++ = ucReversed;
			}
		}

		// delay
		WAITACK();
	}

	return TRUE;
}

// rx 32K block
BOOL CRio::Rx32KBlock( void* pv, UINT uiPos32KBlock )
{
	int iRetry = 0;
	while( iRetry < MAX_RETRY )
	{
		if ( Rx32KBlockRetry(pv, uiPos32KBlock) )
			return TRUE;

		DELAY( CLOCK_SECOND );
		++iRetry;
	}

	ERRORSTR( "too many retries for Rx32KBlock()\n" );

	return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
// operations
BOOL CRio::TxDirectory( void )
{
	// if not set
	if ( !m_bSet )
		return FALSE;

	// create zero initialized temp block
	UCHAR* paucBlock;
	ZERONEWBLOCK( paucBlock );

	// directory header
	CDirHeader& cDirHeader = m_cDirBlock.m_cDirHeader;
	// update directory header time
	cDirHeader.m_lTimeLastUpdate = time( NULL );
	// create checksums (note: second checksum needs to be calculated first as
	//	this makes up part of the first checksum)
	cDirHeader.m_usChecksum2 = CalculateChecksum2();
	cDirHeader.m_usChecksum1 = CalculateChecksum1();

	// copy directory block to temp block
	memcpy( paucBlock, &m_cDirBlock, sizeof(m_cDirBlock) );

	// send block
	BOOL bResult = Tx32KBlock( paucBlock, 0, 0, 0 );
	if ( bResult )
	{
		// IO outro
		IOOutro();
	}

	// release temp block
	DELETEBLOCK( paucBlock );

	return bResult;
}

BOOL CRio::RxDirectory( void )
{
	// if not set
	if ( !m_bSet )
		return FALSE;

	// create temp block
	UCHAR* paucBlock;
	NEWBLOCK( paucBlock );

	// get block
	BOOL bResult = Rx32KBlock( paucBlock, 0 );
	if ( bResult )
	{
		// io outro
		IOOutro();

		// store directory
		memcpy( &m_cDirBlock, paucBlock, sizeof(m_cDirBlock) );

		// validate checksums
		USHORT usChecksum1 = (USHORT)CalculateChecksum1();
		USHORT usChecksum2 = (USHORT)CalculateChecksum2();
		if (
			usChecksum1 != m_cDirBlock.m_cDirHeader.m_usChecksum1 ||
			usChecksum2 != m_cDirBlock.m_cDirHeader.m_usChecksum2
		)
			ERRORSTR( "invalid directory header checksum, recommend initializing\n" );
	}

	// release temp block
	DELETEBLOCK( paucBlock );

	return bResult;
}

BOOL CRio::Initialize( void )
{
	// if not set
	if ( !m_bSet )
		return FALSE;

	// init directory header
	memset( &m_cDirBlock.m_cDirHeader, 0, sizeof(m_cDirBlock.m_cDirHeader) );
	CDirHeader& cDirHeader = m_cDirBlock.m_cDirHeader;
	cDirHeader.m_usCount32KBlockAvailable = 1024;
	cDirHeader.m_usCount32KBlockRemaining = 1024;
	// init directory entries
	memset( &m_cDirBlock.m_acDirEntry, 0, sizeof(m_cDirBlock.m_acDirEntry) );
	// init block used flags
	memset( &m_cDirBlock.m_auc32KBlockUsed, ID_32KBLOCK_FREE, sizeof(m_cDirBlock.m_auc32KBlockUsed) );
	// init FAT
	memset( &m_cDirBlock.m_ausFAT, 0, sizeof(m_cDirBlock.m_ausFAT) );

	return TRUE;
}

BOOL CRio::TxFile( char* pszPathFile, BOOL bVerbose )
{
	// if not set
	if ( !m_bSet )
		return FALSE;

	// directory header
	CDirHeader& cDirHeader = m_cDirBlock.m_cDirHeader;

	// if no enough room in directory for entry
	if ( cDirHeader.m_usCountEntry >= CRIO_MAX_DIRENTRY )
	{
		ERRORSTR( "maximum number of directory entries (%d) already in use\n", CRIO_MAX_DIRENTRY );
		return FALSE;
	}

	// open file for read
	FILE* fpFile = fopen( pszPathFile, "rb" );
	if ( !fpFile )
	{
		ERRORSTR( "unable to open '%s' for read\n", pszPathFile );
		return FALSE;
	}

	// get fileinfo
	struct stat sStat;
	if ( stat(pszPathFile, &sStat) )
	{
		ERRORSTR( "stat() failed for '%s'\n", pszPathFile );
		fclose( fpFile );
		return FALSE;
	}

	// determine if enough room to store file
	USHORT usCount32KBlock = (sStat.st_size/CRIO_SIZE_32KBLOCK) +
		(sStat.st_size % CRIO_SIZE_32KBLOCK ? 1 : 0);
	if ( usCount32KBlock > cDirHeader.m_usCount32KBlockRemaining )
	{
		ERRORSTR( "not enough memory on device to upload '%s'\n", pszPathFile );
		fclose( fpFile );
		return FALSE;
	}

	// get first four bytes of file which determine mp3 properties
	UCHAR aucProperty[ 4 ];
	if ( fread(aucProperty, sizeof(aucProperty), 1, fpFile) < 1 )
	{
		ERRORSTR( "error reading properties from file '%s'\n", pszPathFile );
		fclose( fpFile );
		return FALSE;
	}
	rewind( fpFile );

	// point to directory entry that will be used
	CDirEntry& cDirEntry = m_cDirBlock.m_acDirEntry[ cDirHeader.m_usCountEntry ];

	// update directory header
	if ( !cDirHeader.m_usCountEntry )
	{
		// adjust for dir block
		++cDirHeader.m_usCount32KBlockUsed;
		--cDirHeader.m_usCount32KBlockRemaining;

		// mark first block as used by directory
		m_cDirBlock.m_auc32KBlockUsed[ 0 ] = ID_32KBLOCK_USED;

		// first entry in FAT used by directory
		m_cDirBlock.m_ausFAT[ 0 ] = 0;
	}
	++cDirHeader.m_usCountEntry;
	cDirHeader.m_usCount32KBlockUsed += usCount32KBlock;
	cDirHeader.m_usCount32KBlockRemaining -= usCount32KBlock;

	// find first free 32K block
	USHORT usPos32KBlockFree = FindFirstFree32KBlock();
	if ( usPos32KBlockFree == 0xffff )
	{
		ERRORSTR( "no free 32K blocks available, initialization recommended\n" );
		fclose( fpFile );
		return FALSE;
	}

	// update directory entry
	memset( &cDirEntry, 0, sizeof(CDirEntry) );
	cDirEntry.m_usPos32KBlock = usPos32KBlockFree;
	cDirEntry.m_usCount32KBlock = usCount32KBlock;
	cDirEntry.m_usSize32KMod = sStat.st_size % CRIO_SIZE_32KBLOCK;
	cDirEntry.m_lSize = sStat.st_size;
	cDirEntry.m_lTimeUpload = time( NULL );
	cDirEntry.m_aucProperty[ 0 ] = aucProperty[ 3 ];
	cDirEntry.m_aucProperty[ 1 ] = aucProperty[ 2 ];
	cDirEntry.m_aucProperty[ 2 ] = aucProperty[ 1 ];
	cDirEntry.m_aucProperty[ 3 ] = aucProperty[ 0 ];
	strncpy( cDirEntry.m_szName, GetFile(pszPathFile), sizeof(cDirEntry.m_szName) );

	// create zero initialized temp block
	UCHAR* paucBlock;
	ZERONEWBLOCK( paucBlock );

	// default return
	BOOL bResult = FALSE;

	// process all 32K blocks of file
	USHORT usPos32KBlockCurrent = cDirEntry.m_usPos32KBlock;
	USHORT usPos32KBlockPrev = 0xffff;
	USHORT usPos32KBlockEnd = cDirEntry.m_usCount32KBlock;
	USHORT usPos32KBlockNext;
	USHORT usPos32KBlock;
	for( usPos32KBlock=0; usPos32KBlock<usPos32KBlockEnd; ++usPos32KBlock )
	{
		if ( bVerbose )
		{
			INFOSTR( "blocks %-5hd\r", usPos32KBlockEnd-usPos32KBlock );
			fflush( stdout );
		}

		// mark as block used
		m_cDirBlock.m_auc32KBlockUsed[ usPos32KBlockCurrent ] = ID_32KBLOCK_USED;

		// get next block and update FAT
		if ( usPos32KBlock == (usPos32KBlockEnd-1) )
		{
			usPos32KBlockNext = 0xffff;
			m_cDirBlock.m_ausFAT[ usPos32KBlockCurrent ] = 0;
		}
		else
		{
			usPos32KBlockNext = FindFirstFree32KBlock();
			if ( usPos32KBlockNext == 0xffff )
			{
				ERRORSTR( "no free 32K blocks available, initialization recommended\n" );
				break;
			}
			m_cDirBlock.m_ausFAT[ usPos32KBlockCurrent ] = usPos32KBlockNext;
		}

		// determine read size
		UINT uiSize;
		if ( usPos32KBlock == (usPos32KBlockEnd-1) && cDirEntry.m_usSize32KMod )
		{
			uiSize = cDirEntry.m_usSize32KMod;
			memset( paucBlock+uiSize, 0, CRIO_SIZE_32KBLOCK-uiSize );
		}
		else
			uiSize = CRIO_SIZE_32KBLOCK;

		// read block
		if ( fread(paucBlock, uiSize, 1, fpFile) < 1 )
		{
			ERRORSTR( "error reading from file '%s'\n", pszPathFile );
			break;
		}

		// tx block
		if ( !Tx32KBlock(paucBlock, usPos32KBlockCurrent, usPos32KBlockPrev,
			usPos32KBlockNext) )
			break;

		// update current and prev block markers
		usPos32KBlockPrev = usPos32KBlockCurrent;
		usPos32KBlockCurrent = usPos32KBlockNext;
	}

	// if transfer ok
	if ( usPos32KBlock == usPos32KBlockEnd )
	{
		// flag as ok
		bResult = TRUE;

		// IO outro
		IOOutro();
	}

	// release temp block
	DELETEBLOCK( paucBlock );

	// close file
	fclose( fpFile );

	return bResult;
}

BOOL CRio::RxFile( int entryNum, BOOL bVerbose )
{
	FILE* writeFile;
	CDirEntry* entry;
	
	// check to see if entry exists
	if ( m_cDirBlock.m_cDirHeader.m_usCountEntry < entryNum ){
		if( bVerbose )
			ERRORSTR( "Entry %d does not exist.\n", entryNum );
		return FALSE;
	}
  
	entry = &m_cDirBlock.m_acDirEntry[entryNum-1];
	
	if( bVerbose )
		INFOSTR( "grabbing %s...\n", entry->m_szName );

	// open up the local file for writing
	if(!( writeFile = fopen( entry->m_szName, "wb" ) )){
		ERRORSTR( "Could not write %s.\n", entry->m_szName );
		return FALSE;
	}
	
	// create temp block
	UCHAR* paucBlock;
	NEWBLOCK( paucBlock );
	
	int blockNum;
	int lastBlock = entry->m_usPos32KBlock + entry->m_usCount32KBlock - 1;
	
	// fetch the entry, block by block
	for( blockNum = entry->m_usPos32KBlock ; blockNum <= lastBlock ; blockNum++ ){
	  
		// read in the data
		if(! Rx32KBlock( paucBlock, blockNum )){
			if( bVerbose )
				ERRORSTR( "There was an error fetching block %d\n", blockNum );
			return FALSE;
		}

		// write it out to the file
		if( blockNum != lastBlock ){
			fwrite( paucBlock, CRIO_SIZE_32KBLOCK, 1, writeFile );
		}else{
			fwrite( paucBlock, entry->m_usSize32KMod, 1, writeFile );
		}
	}
	
	// cleanup
	IOOutro();
	fclose( writeFile );
	DELETEBLOCK( paucBlock );
}





