///////////////////////////////////////////////////////////////////////////////
//
//	rio.h
//
///////////////////////////////////////////////////////////////////////////////

#ifndef		_RIO_
#define		_RIO_

// dependencies
#include	"std.h"

// max available 32k blocks
#define		CRIO_MAX_32KBLOCK			8192

// size of 32k block
#define		CRIO_SIZE_32KBLOCK			32768

// max directory entries
#define		CRIO_MAX_DIRENTRY			60

///////////////////////////////////////////////////////////////////////////////
// enable byte align structs
#pragma pack( 1 )

// directory header
struct CDirHeader
{
	USHORT m_usCountEntry;
	USHORT m_usCount32KBlockAvailable;
	USHORT m_usCount32KBlockUsed;
	USHORT m_usCount32KBlockRemaining;
	char m_acNotUsed[ 2 ];
	long m_lTimeLastUpdate;
	USHORT m_usChecksum1;
	USHORT m_usChecksum2;
	char m_acNotUsed2[ 512 - 18 ];
};

// directory entry
struct CDirEntry
{
	USHORT m_usPos32KBlock;
	USHORT m_usCount32KBlock;
	USHORT m_usSize32KMod;
	long m_lSize;
	char m_acNotUsed[ 5 ];
	long m_lTimeUpload;
	char m_aucProperty[ 4 ];
	char m_acNotUsed3[ 5 ];
	char m_szName[ 128 - 28 ];
};

// directory block
struct CDirBlock
{
	CDirHeader m_cDirHeader;
	CDirEntry m_acDirEntry[ CRIO_MAX_DIRENTRY ];
	UCHAR m_auc32KBlockUsed[ CRIO_MAX_32KBLOCK ];
	USHORT m_ausFAT[ CRIO_MAX_32KBLOCK ];
};

// end 512 byte block
struct CEnd512ByteBlock
{
	ULONG m_ulPos512ByteBlockNextMult256;
	UCHAR m_ucPos8192KBlockNext1;
	UCHAR m_ucPos8192KBlockNext2;
	ULONG m_ulPos512ByteBlockPrevMult256;
	UCHAR m_ucPos8192KBlockPrev1;
	UCHAR m_ucPos8192KBlockPrev2;
	USHORT m_usChecksum;
	USHORT m_usPos32KBlockPrevMult256;
};

// disable byte align
#pragma pack()

///////////////////////////////////////////////////////////////////////////////
// rio class
class CRio
{
	// directory block
	CDirBlock m_cDirBlock;

	// port constants
	int m_iPortBase;
	int m_iPortData;
	int m_iPortStatus;
	int m_iPortControl;

	// successful setup flag
	BOOL m_bSet;

	UINT FindFirstFree32KBlock( void );
	UINT CalculateChecksum1( void );
	UINT CalculateChecksum2( void );
	BOOL WaitInput( int iValue );
	BOOL WaitAck( void );
	BOOL IOIntro( void );
	BOOL IOOutro( void );

	BOOL Tx32KBlockRetry( void* pv, UINT uiPos32KBlock, UINT uiPos32KBlockPrev,
		UINT uiPos32KBlockNext );
	BOOL Tx32KBlock( void* pv, UINT uiPos32KBlock, UINT uiPos32KBlockPrev,
		UINT uiPos32KBlockNext );
	BOOL Rx32KBlockRetry( void* pv, UINT uiPos32KBlock );
	BOOL Rx32KBlock( void* pv, UINT uiPos32KBlock );

public:

	// set/unset
	void Unset( void );
	BOOL Set( int iPortBase );

	// constructors/destructors
	CRio();
	~CRio();

	// retrieval
	CDirBlock& GetDirectoryBlock( void ) { return m_cDirBlock; }

	// operations
	BOOL TxDirectory( void );
	BOOL RxDirectory( void );
	BOOL Initialize( void );
	BOOL TxFile( char* pszPathFile, BOOL bVerbose );
	BOOL RxFile( int entry, BOOL bVerbose );
};

#endif		// _RIO_
